/*
 * abcread.c
 *
 * Program to extract the contents from an ABC/UFD-DOS filesystem
 */
#include "abcdisk.h"

/* Copy files as given by tree on stdin */
static int
extract_dir_sector(struct disk *disk, const char *dirpath, int sectnr)
{
    char filename[4096], *p;
    unsigned char *sector;
    unsigned char *header;
    int i, hsect, dsect, len;
    int ufd, skip, cleft;
    FILE *f;

    /* 16 directory entries per sector */
    sector = disk->image + (sectnr << 8);

    for (i = 0; i < 16; i++, sector += 16) {
        hsect = (sector[0] << 3) + (sector[1] >> 5);
        hsect <<= disk->fmt->cluster_shift;
        if (!hsect || (sector[0] & sector[1]) == 0xff ||
            sector[4] == 0x00 || sector[4] == 0xff)
            continue;           /* Empty slot */

        header = disk->image + (hsect << 8);

        /* Length of file in sectors, including header block */
        len = sector[2] + (sector[3] << 8);

        /*
         * Early version of ABC-DOS didn't fill in the length field; those
         * versions all had cluster size == 1 so we can just read until the
         * header block runs out...
         */
        if (len == 0)
            len = disk->fmt->sectors;

        ufd = !memcmp(sector + 12, "Ufd", 3);
        p = filename + sprintf(filename, "%s/", dirpath);
        unmangle_filename(p, sector + 4);

        if (ufd)
            make_dir(filename);
        else
            f = fopen(filename, "wb");

        /* Skip the header block for a file, header block and bitmap for UFD */
        skip = 1 + ufd;
        cleft = 0;
        dsect = -1;
        header += 4;

        while (len--) {
            if (cleft) {
                dsect++;
            } else {
                if ((header[0] & header[1]) == 0xff)
                    break;      /* Premature end of data */

                dsect = (header[0] << 3) + (header[1] >> 5);
                dsect <<= disk->fmt->cluster_shift;
                cleft = ((header[1] & 31) + 1) << disk->fmt->cluster_shift;

                header += 2;
            }
            cleft--;

            if (skip) {
                skip--;
            } else if (ufd) {
                extract_dir_sector(disk, filename, dsect);
            } else {
                fwrite(disk->image + (dsect << 8) + 3, 1, 253, f);
            }
        }

        if (!ufd)
            fclose(f);
    }

    return 0;
}

int main(int argc, char *argv[])
{
    struct disk *disk;
    const struct diskfmt *fmt;
    FILE *f;
    int i;

    setlocale(LC_ALL, "");

    program = argv[0];

    if (argc != 4 || !(fmt = get_format(argv[1]))) {
	fprintf(stderr, "abcdisk %s\n", VERSION);
        fprintf(stderr, "Usage: %s format file outdir\n", program);
        show_formats(stderr);
        exit(1);
    }

    disk = allocate_disk(fmt);

    f = !strcmp(argv[2], "-") ? stdin : fopen(argv[2], "rb");
    fread(disk->image, 1, fmt->sectors << 8, f);
    if (f != stdin)
        fclose(f);

    /* Extract files */
    for (i = 0; i < 16; i++)
        extract_dir_sector(disk, argv[3], i + 16);

    return 0;
}
