/*
 * abcwrite.c
 *
 * Program to build a populated ABC-DOS hard disk image
 */
#include "abcdisk.h"

/*
 * Find a free cluster in the bitmap; return the cluster number
 */
static int get_cluster(struct disk *disk, unsigned int cluster_hint)
{
    unsigned int nclust, i, x, nc, best, bnc;

    best = -1;
    bnc = 0;

    nclust = disk->fmt->sectors << disk->fmt->cluster_shift;

    for (i = 0; i < nclust; i++) {
        if (!is_used(disk, i)) {
            /* How long is this? */
            for (x = i, nc = 0; i < nclust && nc < cluster_hint; i++) {
                if (is_used(disk, i))
                    break;
                nc++;
            }
            if (nc == cluster_hint) {
                /* Found as big as we wanted */
                mark_used(disk, x);
                return x;
            } else if (nc > bnc) {
                best = x;
                bnc = nc;
            }
        }
    }

    if (bnc) {
        /* We found *something* at least... */
        mark_used(disk, best);
        return best;
    }

    fprintf(stderr, "%s: output disk full\n", program);
    exit(1);
}

/*
 * Get the contents of a file, and guess whether or not it's a text file
 */

static unsigned char *read_file(const char *filename, int *sizep)
{
    FILE *f = fopen(filename, "rb");
    unsigned char buffer[256];
    int i, n;
    int size = 0;
    int binary = 0;
    unsigned char *data, *dp;
    int blocks;

    if (!f)
        return NULL;

    while ((n = fread(buffer, 1, sizeof buffer, f))) {
        if (!binary)
            for (i = 0; i < n; i++) {
                if ((buffer[i] & 0x80) || buffer[i] == 0x00
                    || buffer[i] == 0x03) {
                    binary = 1;
                    break;
                }
            }

        size += n;
    }

    rewind(f);

    if (binary) {
        blocks = (size + 252) / 253;
        data = xmalloc(blocks * 253);
        fread(data, size, 1, f);
        memset(data + size, 0, blocks * 253 - size);
        size = blocks * 253;
    } else {
        int blocks = (size + 251) / 252 + 1;
        data = xmalloc(blocks * 253);

        dp = data;
        while (memset(buffer, 0, 253), (n = fread(buffer, 1, 252, f))) {
            for (i = 0; i < n; i++) {
                if (buffer[i] == '\n')
                    buffer[i] = '\r';
            }
            buffer[n] = '\x03';

            memcpy(dp, buffer, 253);
            dp += 253;
        }

        /* Ending block */
        memset(buffer, 0, 6);
        buffer[6] = '\x03';
        memcpy(dp, buffer, 253);
        dp += 253;

        size = dp - data;
    }

    fclose(f);

    *sizep = size;

    return data;
}

/*
 * Create a file object and return a pointer to the header block.
 */
static unsigned char *create_file(struct directory *dir,
                                  const unsigned char *mangled_name,
				  unsigned char *data,
                                  int size, int ufd)
{
    struct disk *disk = dir->disk;
    int nblocks, cluster, secperclust, clusters;
    int dirs, dirp;             /* Directory (sector,position) */
    unsigned char *dirent;
    int blockno, cleft;
    unsigned char *header, *hdata, *dp;

    if (!data)
        return NULL;            /* No data, not even an empty buffer */

    nblocks = (size + 252) / 253 + 1;
    secperclust = 1 << disk->fmt->cluster_shift;
    clusters = (nblocks + secperclust - 1) >> disk->fmt->cluster_shift;

    /* Find and create directory entry */
    dirp = -1;
    for (dirs = 0; dirs < disk->fmt->dirsecs; dirs++) {
        if (dir->bitmap[0xEF + dirs] < 0x10) {
            dirp = dir->bitmap[0xEF + dirs]++;
            break;
        }
    }

    if (dirs >= disk->fmt->dirsecs) {
        /* Directory full */
        fprintf(stderr, "%s: directory full\n", program);
        return NULL;
    }

    dirent = dir->dirsec + dirs * 256 + dirp * 16;

    memcpy(dirent + 4, mangled_name, 11);

    /* Allocate first cluster */
    cluster = get_cluster(disk, clusters--);
    if (cluster < 0) {
        fprintf(stderr, "get_cluster = %d\n", cluster);
        return NULL;
    }

    dirent[0] = cluster >> 3;
    dirent[1] = cluster << 5;
    dirent[2] = nblocks & 0xFF;
    dirent[3] = nblocks >> 8;

    /* Create "file header" */
    dp = header = disk->image + (cluster << (disk->fmt->cluster_shift + 8));
    cleft = secperclust - 1;    /* Left in the first cluster */

    memset(hdata = header, 0, 256);

    *hdata++ = (dirp << 4) + dirs;
    *hdata++ = 0;               /* Header = block 0 */
    *hdata++ = 0;               /* Header = block 0 */
    *hdata++ = 0xFF;
    /* Starting extent has one cluster in it */
    hdata[0] = cluster >> 3;
    hdata[1] = (cluster << 5) + 0;      /* Bits[4:0] = extend size minus one */

    /* Actually copy the file */
    blockno = 0;
    while (size) {
        int nbytes = (size > 253) ? 253 : size;

        blockno++;

        if (cleft) {
            dp += 256;          /* Subsequent sector */
            cleft--;
        } else {
            cluster = get_cluster(disk, clusters--);
            if (cluster < 0) {
                fprintf(stderr, "get_cluster = %d\n", cluster);
                return NULL;
            }
            dp = disk->image + (cluster << (disk->fmt->cluster_shift + 8));
            cleft = secperclust - 1;
            if (cluster == ((hdata[0] << 3) + (hdata[1] >> 5) + 1) &&
                (hdata[1] & 31) < 31) {
                hdata[1]++;
            } else {
                hdata += 2;
                hdata[0] = cluster >> 3;
                hdata[1] = (cluster << 5) + 0;
            }
        }

        if (ufd && blockno > 1) {
            memset(dp, 0xff, 256);
        } else {
            dp[0] = (dirp << 4) + dirs;
            dp[1] = blockno & 0xFF;
            dp[2] = blockno >> 8;
            memcpy(dp + 3, data, 253);
            data += 253;
        }

        size -= nbytes;
    }

    /* Terminate the hdata */
    hdata += 2;
    hdata[0] = 0xFF;
    hdata[1] = 0xFF;

    return header;
}

/*
 * Copy a file
 */
static int copy_file(struct directory *dir, const char *filename)
{
    int size;
    unsigned char mangled_name[12];
    unsigned char *data;
    unsigned char *header;

    data = read_file(filename, &size);

    mangle_name(mangled_name, filename);

    header = create_file(dir, mangled_name, data, size, 0);
    free(data);
    return header ? 0 : -1;
}

/*
 * Create a UFD
 */
static struct directory *make_ufd(struct directory *dir, const char *filename)
{
    unsigned char mangled_name[12];
    unsigned char bitmap[253];
    unsigned char *header;
    struct directory *ufd;
    int dirs_block = dir->block ? dir->block + 2 : 0;

    ufd = xmalloc(sizeof(struct directory));
    memset(ufd, 0, sizeof *ufd);
    ufd->disk = dir->disk;

    mangle_name(mangled_name, filename);
    memcpy(mangled_name + 8, "Ufd", 4);

    memset(bitmap, 0, sizeof bitmap);
    bitmap[0] = dirs_block & 0xFF;
    bitmap[1] = dirs_block >> 8;

    header = create_file(dir, mangled_name, bitmap,
                         (dir->disk->fmt->dirsecs + 1) * 253, 1);
    if (!header) {
        fprintf(stderr, "%s: header returns NULL\n", program);
        return NULL;
    }

    if ((header[6] & header[7]) != 0xff) {
        fprintf(stderr, "%s: discontiguous UFD (disk too fragmented)\n",
                program);
        return NULL;
    }

    ufd->bitmap = header + 256;
    ufd->dirsec = header + 512;
    ufd->block = (header - dir->disk->image) >> 8;

    /* The UFD bitmap contains two copies of the self-pointer... */
    ufd->bitmap[5] = ufd->bitmap[0];

    return ufd;
}

/* Copy files as given by tree on stdin */
static int get_files(struct directory *ufd, const char *dirpath)
{
    struct stat st;
    DIR *d;
    struct dirent *de;
    struct directory *subdir;
    size_t dirlen, filelen;
    char *filename;

    d = opendir(dirpath);
    if (!d) {
        /* Directory error or not a directory */

        int err = errno;
        int rv;

        /* Is it a single file? */
        rv = copy_file(ufd, dirpath);
        if (rv) {
            errno = err;
            perror(dirpath);
        }
        return rv;
    }

    dirlen = strlen(dirpath);

    while ((de = readdir(d))) {
        if (de->d_name[0] == '.')
            continue;

        filelen = strlen(de->d_name);
        filename = xmalloc(dirlen + filelen + 2);
        memcpy(filename, dirpath, dirlen);
        filename[dirlen] = '/';
        memcpy(filename + dirlen + 1, de->d_name, filelen + 1);

        if (stat(filename, &st)) {
            perror(filename);
            continue;
        }
        if (S_ISDIR(st.st_mode)) {
            subdir = make_ufd(ufd, filename);
            if (subdir)
                get_files(subdir, filename);
            else
                fprintf(stderr, "%s: make_ufd(%s) failed\n", program, filename);
        } else {
            copy_file(ufd, filename);
        }
        free(filename);
    }
    closedir(d);

    return 0;
}

int main(int argc, char *argv[])
{
    struct disk *disk;
    const struct diskfmt *fmt;
    FILE *f;

    setlocale(LC_ALL, "");

    program = argv[0];

    if (argc != 4 || !(fmt = get_format(argv[1]))) {
	fprintf(stderr, "abcdisk %s\n", VERSION);
        fprintf(stderr, "Usage: %s format indir outfile\n", argv[0]);
        show_formats(stderr);
        return 1;
    }

    disk = allocate_disk(fmt);

    /* Install files */
    if (get_files(&disk->mfd, argv[2]))
        return 1;               /* No files to be had */

    /* Create backup MFD if it exists */
    if (fmt->dirsec[1]) {
        memcpy(disk->image + (fmt->dirsec[1] << 8),
               disk->image + (fmt->dirsec[0] << 8), fmt->dirsecs << 8);
    }

    /* Finally, write out image */
    f = !strcmp(argv[3], "-") ? stdout : fopen(argv[3], "wb");
    if (!f) {
        perror(argv[3]);
        return 1;
    }
    fwrite(disk->image, 1, fmt->sectors << 8, f);
    if (f != stdout)
        fclose(f);

    return 0;
}
